/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamskin.h"
#include <gdam/component/gdambuffermanager.h>
#include <gtk/gtkprogressbar.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwindow.h>
#include <gdk/gdk.h>
#include <gdam/component/gdamutil.h>
#include "gdamskinutil.h"
#include <gdam/gdamplugin.h>
#include <gdam/gdamsearchpath.h>
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkfeatures.h>
#include <gtk/gtklabel.h>
#include <gnome-xml/tree.h>
#include <stdlib.h>

static void gdam_skin_destroy(GtkObject* object);

static GtkObjectClass* parent_class = NULL;
static void signal_name_hacks(const char* signal_name,
                              GtkObject*  object);

static gboolean gdam_skin_build_new_model(GdamSkin*      skin,
				          GdamChannel*   channel,
				          int            num_options,
	                                  char**         options)
{
	GdamModel* model;
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(GTK_OBJECT(skin)->klass);
	if (skin_class->model_type == GTK_TYPE_NONE) {
		gdam_debug("skin class %s didn't have build_new_model method, "
			"but didn't have model type either",
			gtk_type_name(GTK_OBJECT(skin)->klass->type));
		return TRUE;
	}

	model = GDAM_MODEL(gtk_type_new(skin_class->model_type));
	g_return_val_if_fail(model != NULL, FALSE);
	model->channel = channel;
	skin->model = model;
	return TRUE;
}
static void gdam_skin_attach_to_model(GdamView*     view,
                                      GdamModel*    model)
{
	GdamSkin* skin = GDAM_SKIN(view);
	g_return_if_fail(skin->model == NULL);
	g_return_if_fail(model != NULL);
	skin->model = model;
	gtk_object_ref(GTK_OBJECT(model));
}

static const char* xml_path = 
	 "~/.gdam/skins"
	 ":."
	 ":/etc/gdam/skins"
	 ":" GDAM_SKIN_PREFIX
	 ":/usr/local/lib/gdam/skins"
	 ":/usr/lib/gdam/skins";

static gboolean has_connect_handler       (const gchar*       handler_name,
                                           GtkType            type);
static gboolean skin_generic_connect      (const gchar*       handler,
                                           GtkObject*         object,
				           const gchar*       signal_name,
				           GdamSkin*          skin);

static gboolean gdam_skin_initialize_main_widget
				          (GdamSkin*          skin,
				           const char*        skin_filename);
static gboolean gdam_skin_initialize_widgets(GdamSkin* skin);

static gboolean gdam_skin_make_xml        (GdamView*          view,
                                           xmlNode*           parent,
				           GdamXmlInfo*       xml_info);
static gboolean gdam_skin_parse_xml       (GdamView*          view,
                                           xmlNode*           parent,
					   GdamXmlInfo*       xml_info);

static gboolean gdam_skin_toplevel_handler(GdamSkin*          skin,
				           GtkObject*         object,
				           const char*        signal_name,
				           int                num_handler_names,
				           char**             handler_names)
{
	gtk_signal_connect_object(GTK_OBJECT(skin),
	                          "destroy",
			          GTK_SIGNAL_FUNC(gtk_widget_destroy),
				  object);
	return TRUE;
}

/*=====================================================================
 *                     ``gdam_skin_launch''
 *
 * Define a button or whatever to launch a new skin
 * when clicked.
 */
typedef struct _LaunchInfo {

	/* The skin which is doing the launching. */
	GdamSkin*     launcher_skin;

	/* The name of the type of skin to instantiate. */
	char*         model_name;

	/* The xml file to use for the skin. */
	char*         skin_name;

	char*	      unique_id;

	/* A collection of arguments to the instantiation. */
	int           num_options;
	char**        options;
} LaunchInfo;

static void launch_widget(GdamSkin* skin)
{
	GtkWidget* window = NULL;
	if (skin->main_widget == NULL) {
		g_warning("launch widget failed[%s]: main_widget was NULL",
			gtk_type_name(GTK_OBJECT(skin)->klass->type));
		return;
	}
	if (! GTK_WIDGET_TOPLEVEL(skin->main_widget)) {
		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);
		gdam_set_container(GTK_CONTAINER(window), skin->main_widget);
	}

	if (! GTK_WIDGET_VISIBLE(skin->main_widget))
		gtk_widget_show(skin->main_widget);
	if (window)
		gtk_widget_show(window);
}

/*
 * Instantiate a new skin from a button callback.
 * Returns the skin b/c it is a helpful utility function
 * as well.
 */
static GdamSkin*
gdam_skin_launch_callback(GtkWidget* widget, LaunchInfo* launchinfo)
{
	GdamModel* model = NULL;
	GdamSkin* skin;
	GdamChannel* channel = launchinfo->launcher_skin->channel;
	GdamToplevel* toplevel = gdam_toplevel_get_global(channel);
	skin = gdam_skin_new_by_type_name(launchinfo->model_name,
			                  launchinfo->skin_name, 
					  launchinfo->num_options,
					  launchinfo->options,
			                  channel,
			                  &model);
	if (!skin) {
		g_warning("error building %s/%s",
			  launchinfo->model_name, 
			  launchinfo->skin_name);
		return NULL;
	}

	if (model)
		gdam_toplevel_add_child(toplevel, model);
	gdam_skin_connect_toplevel_notify(skin, toplevel);
	launch_widget(skin);
	return skin;
}

static gboolean gdam_skin_launch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	char* skin_type;
	char* skin_filename;
	LaunchInfo* launchinfo;
	if ((skin_type =     handler_names[1]) == NULL
	 || (skin_filename = handler_names[2]) == NULL) {
		g_warning("too few arguments to %s", handler_names[0]);
		return TRUE;
	}

	handler_names[1] = handler_names[2] = NULL;

	launchinfo = g_new(LaunchInfo, 1);
	launchinfo->launcher_skin = skin;
	launchinfo->model_name = skin_type;
	launchinfo->skin_name = skin_filename;

	launchinfo->num_options = num_handler_names - 3;
	launchinfo->options = gdam_strdupv(handler_names + 3);
	gtk_signal_connect(object,
	                   "clicked",
	                   GTK_SIGNAL_FUNC(gdam_skin_launch_callback),
			   launchinfo);
	return TRUE;
}

/*=====================================================================
 *                   ``gdam_skin_launch_unique''
 *                  ``gdam_skin_raise_or_launch''
 *
 * Same as `gdam_skin_launch', except that if the skin already
 * exists:
 *
 *      launch_unique       will toggle it from hidden to shown and back
 *      raise_or_launch     will raise it to the front.
 */

static GHashTable* unique_skins = NULL;
static GHashTable* skin_to_name = NULL;

GdamSkin* gdam_skin_try_get_unique(const char* skin_type)
{
	gpointer ptr;
	if (unique_skins == NULL)
		return NULL;
	ptr = g_hash_table_lookup(unique_skins, skin_type);
	if (ptr == NULL)
		return NULL;
	return GDAM_SKIN(ptr);
}

static void unique_launched_skin_destroy_handler(GdamSkin* skin)
{
	char* name = g_hash_table_lookup(skin_to_name, skin);
	g_return_if_fail(name != NULL);
	g_hash_table_remove(unique_skins, name);
	g_hash_table_remove(skin_to_name, skin);
	g_free(name);
}

static void
gdam_skin_launch_unique_callback(GtkWidget* widget, LaunchInfo* launchinfo)
{
	
	GdamSkin* skin;
	char* name;
	if (unique_skins == NULL) {
		unique_skins = g_hash_table_new(g_str_hash, g_str_equal);
		skin_to_name = g_hash_table_new(NULL, NULL);
	}
	skin = g_hash_table_lookup(unique_skins, launchinfo->unique_id);
	if (skin != NULL) {
		GtkWidget* widget = GDAM_SKIN(skin)->main_widget;
		if (widget == NULL) {
			g_warning("skin was no widget to raise");
			return;
		}
		if (GTK_WIDGET_VISIBLE(widget)) {
			gtk_widget_hide(widget);
		} else {
			gtk_widget_show(widget);
		}
		return;
	}

	skin = gdam_skin_launch_callback(widget, launchinfo);
	name = g_strdup(launchinfo->unique_id);
	g_hash_table_insert(unique_skins, name, skin);
	g_hash_table_insert(skin_to_name, skin, name);
	gtk_signal_connect(GTK_OBJECT(skin),
	                   "destroy", 
		   GTK_SIGNAL_FUNC(unique_launched_skin_destroy_handler),
			   NULL);
}

static void
gdam_skin_raise_or_launch(GtkWidget* widget, LaunchInfo* launchinfo)
{
	
	GdamSkin* skin;
	char* name;
	if (unique_skins == NULL) {
		unique_skins = g_hash_table_new(g_str_hash, g_str_equal);
		skin_to_name = g_hash_table_new(NULL, NULL);
	}
	skin = g_hash_table_lookup(unique_skins, launchinfo->model_name);
	if (skin != NULL) {
		GtkWidget* widget = GDAM_SKIN(skin)->main_widget;
		GdamSkinClass* klass;
		if (widget == NULL) {
			g_warning("skin was no widget to raise");
			return;
		}
		if (!GTK_WIDGET_VISIBLE(widget))
			gtk_widget_show(widget);
		while (widget->parent)
			widget = widget->parent;
		if (!GTK_WIDGET_VISIBLE(widget))
			gtk_widget_show(widget);
		gdk_window_raise(widget->window);
		klass = GDAM_SKIN_CLASS(GTK_OBJECT(skin)->klass);
		if (klass->evaluate_options != NULL)
			(*klass->evaluate_options)(skin, 
					           launchinfo->num_options,
					           launchinfo->options);
		return;
	}

	skin = gdam_skin_launch_callback(widget, launchinfo);
	name = g_strdup(launchinfo->unique_id);
	g_hash_table_insert(unique_skins, name, skin);
	g_hash_table_insert(skin_to_name, skin, name);
	gtk_signal_connect(GTK_OBJECT(skin),
	                   "destroy", 
		   GTK_SIGNAL_FUNC(unique_launched_skin_destroy_handler),
			   NULL);
}

#define gdam_skin_raise_or_launch_handler gdam_skin_launch_unique_handler
static gboolean gdam_skin_launch_unique_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	char* skin_type;
	char* skin_filename;
	LaunchInfo* launchinfo;
	GtkSignalFunc signal_func;
	if ((skin_type =     handler_names[1]) == NULL
	 || (skin_filename = handler_names[2]) == NULL) {
		g_warning("too few arguments to %s", handler_names[0]);
		return TRUE;
	}

	handler_names[1] = handler_names[2] = NULL;

	launchinfo = g_new(LaunchInfo, 1);
	launchinfo->launcher_skin = skin;
	launchinfo->model_name = skin_type;
	launchinfo->skin_name = skin_filename;
	launchinfo->unique_id = handler_names[3];
	
	if (handler_names[3] == NULL) {
		launchinfo->unique_id = g_strdup(skin_type);
		launchinfo->num_options = 0;
		launchinfo->options = NULL;
	} else {
		handler_names[3] = NULL;
		launchinfo->num_options = num_handler_names - 4;
		launchinfo->options = gdam_strdupv(handler_names + 4);
	}
	if (strcmp(handler_names[0], "gdam_skin_launch_unique") == 0) {
		signal_func = GTK_SIGNAL_FUNC(gdam_skin_launch_unique_callback);
	} else {
		signal_func = GTK_SIGNAL_FUNC(gdam_skin_raise_or_launch);
	}
	gtk_signal_connect(object, "clicked", signal_func, launchinfo);
	return TRUE;
}

/*=====================================================================
 *              ``gdam_skin_spin_button_display_beatstyle''
 *
 * Take advantage of a feature in gtk1.3 to render the spinbutton
 * label text differently.
 */

#if defined(GTK_CHECK_VERSION)
#if GTK_CHECK_VERSION(1,3,0)
#define MAX_TEXT_LENGTH	256

static gint spin_button_beatstyle_output(GtkSpinButton* spin_button)
{
	gchar buf[MAX_TEXT_LENGTH];

	gfloat value = spin_button->adjustment->value;
	if (value < 0.0) {
		sprintf (buf, "1/%0.*f",
		         spin_button->digits,
			 spin_button->adjustment->value);
	} else {
		sprintf (buf, "%0.*f",
		         spin_button->digits,
			 spin_button->adjustment->value);
	}
	if (strcmp (buf, gtk_entry_get_text (GTK_ENTRY (spin_button))))
		gtk_entry_set_text (GTK_ENTRY (spin_button), buf);
	return TRUE;
}
#endif
#endif

static gboolean gdam_skin_spin_button_display_beatstyle_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	g_return_val_if_fail(GTK_IS_SPIN_BUTTON(object), TRUE);
#if defined(GTK_CHECK_VERSION)
#if GTK_CHECK_VERSION(1,3,0)
	gtk_signal_connect(object, "output",
			   GTK_SIGNAL_FUNC(spin_button_beatstyle_output),
			   NULL);
#endif
#endif
	return TRUE;
}

/*
 *                    ``gdam_skin_destroy''
 *
 * Make a button that will close the skin it is contained
 * in.
 */
static gboolean gdam_skin_destroy_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
	                          GTK_SIGNAL_FUNC(gtk_object_destroy),
				  (GtkObject*) skin);
	return TRUE;
}

static void destroy_skin_top(GdamSkin* skin)
{
	GtkWidget* main_widget = skin->main_widget;
	GtkWidget* top;
	for (top = main_widget; top->parent != NULL; top = top->parent)
		;
	if (top == main_widget) {
		gtk_object_destroy(GTK_OBJECT(main_widget));
	} else {
		gtk_object_destroy(GTK_OBJECT(main_widget));
		gtk_object_destroy(GTK_OBJECT(top));
	}
}

static gboolean gdam_skin_destroy_top_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
	                          GTK_SIGNAL_FUNC(destroy_skin_top),
				  (GtkObject*) skin);
	return TRUE;
}


/*
 *                         ``gdam_skin_adjustment''
 *
 * Attach an adjustment to a GdamArg of the model.
 */
typedef struct _GdamSkinAdjustment GdamSkinAdjustment;

struct _GdamSkinAdjustment {
	GtkAdjustment*		adjustment;
	GtkWidget*		adj_widget;
	GtkObject*		object;
	gboolean		inverted;
	char*			arg_name;
	guint			arg_connect_id;
	guint			adj_connect_id;
	guint			obj_destroy_id;
	guint			adj_destroy_id;
	gboolean		is_int;

	void                  (*do_widget_change)(GdamSkinAdjustment*,
	                                          gdouble new_val);
};

static void free_skin_adjustment(GdamSkinAdjustment* skin_adj)
{
	if (skin_adj->object != NULL) {
		gdam_arg_disconnect(skin_adj->object, skin_adj->arg_connect_id);
		gtk_signal_disconnect(skin_adj->object,
		                      skin_adj->obj_destroy_id);
		skin_adj->object = NULL;
	}
	if (skin_adj->adjustment != NULL) {
		if (skin_adj->adj_connect_id)
			gtk_signal_disconnect(GTK_OBJECT(skin_adj->adjustment),
					      skin_adj->adj_connect_id);
		gtk_signal_disconnect(GTK_OBJECT(skin_adj->adjustment),
		                      skin_adj->adj_destroy_id);
		skin_adj->adjustment = NULL;
	}
	g_free(skin_adj->arg_name);
	g_free(skin_adj);
}

static void skin_adj_arg_changed(gpointer       info_ptr,
                                 GdamArg*       arg,
				 GtkObject*     object)
{
	GdamSkinAdjustment* skin_adj = info_ptr;
	GtkAdjustment* adjustment = skin_adj->adjustment;
	gdouble value;
	if (arg->type == GDAM_ARG_TYPE_INT)
		value = GDAM_ARG_INT(arg);
	else if (arg->type == GDAM_ARG_TYPE_DOUBLE)
		value = GDAM_ARG_DOUBLE(arg);
	else {
		g_warning("type mismatch to adjustment: %s", arg->name);
		return;
	}
	if (skin_adj->inverted) {
		value = adjustment->lower - value
	              + (adjustment->upper - adjustment->page_size);
	}
	if (skin_adj->do_widget_change != NULL)
		(*skin_adj->do_widget_change)(skin_adj, value);
	else
		gtk_adjustment_set_value(skin_adj->adjustment, value);
}
static void adj_object_destroyed(GdamSkinAdjustment* skin_adj)
{
	skin_adj->object = NULL;
	free_skin_adjustment(skin_adj);
}
static void adj_adjust_destroyed(GdamSkinAdjustment* skin_adj)
{
	skin_adj->adjustment = NULL;
	free_skin_adjustment(skin_adj);
}

static gdouble gdam_skin_adjustment_get_value(GdamSkinAdjustment* skin_adj,
                                              gdouble             adj_value)
{
	GtkAdjustment* adj = skin_adj->adjustment;
	gdouble value = adj_value;
	if (skin_adj->inverted) {
		value = adj->lower - value
	              + (adj->upper - adj->page_size);
	}
	return value;
}
static void adj_value_changed(GdamSkinAdjustment* skin_adj)
{
	GdamArg arg;
	GtkAdjustment* adj;
	gdouble value;
	int type;
	arg.name = skin_adj->arg_name;
	adj = skin_adj->adjustment;
	type = gdam_arg_query_type(skin_adj->object->klass->type, arg.name);
	value = gdam_skin_adjustment_get_value(skin_adj, adj->value);
	if (type == GDAM_ARG_TYPE_INT) {
		GDAM_ARG_INT(&arg) = (int)value;
		arg.type = GDAM_ARG_TYPE_INT;
	} else {
		GDAM_ARG_DOUBLE(&arg) = value;
		arg.type = GDAM_ARG_TYPE_DOUBLE;
	}
	gdam_arg_set_object_arg(skin_adj->object, &arg);
}

static void update_progress_bar(GdamSkinAdjustment* skin_adj, gdouble val)
{
	val = CLAMP(val, 0.0, 1.0);
	gtk_progress_bar_update(GTK_PROGRESS_BAR(skin_adj->adj_widget), val);
}

static GdamSkinAdjustment* gdam_skin_adjustment_new(GtkObject* object,
						    GtkWidget* adj_widget,
						    const char* name,
						    gboolean inverted)
{
	GdamSkinAdjustment* skin_adj;
	GtkAdjustment* adj;
	g_return_val_if_fail(adj_widget != NULL, NULL);
	adj = util_get_widget_adjustment(adj_widget);
	g_return_val_if_fail(adj != NULL, NULL);
	skin_adj = g_new(GdamSkinAdjustment, 1);
	skin_adj->adj_widget = adj_widget;
	skin_adj->adjustment = adj;
	skin_adj->object = object;
	skin_adj->inverted = inverted;
	skin_adj->arg_name = g_strdup(name);
	skin_adj->arg_connect_id = gdam_arg_connect(object,
	                                            name,
						    skin_adj_arg_changed,
						    skin_adj);
	skin_adj->is_int = FALSE;
	skin_adj->do_widget_change = NULL;
	if (GTK_IS_PROGRESS_BAR(adj_widget)) {
		skin_adj->do_widget_change = update_progress_bar;
		adj = NULL;
	}
	skin_adj->obj_destroy_id
		= gtk_signal_connect_object
		                   (object,
	                            "destroy",
				    GTK_SIGNAL_FUNC(adj_object_destroyed),
				    (GtkObject*) skin_adj);
	skin_adj->adj_destroy_id
		= gtk_signal_connect_object
		                   (GTK_OBJECT(skin_adj->adjustment),
	                            "destroy",
				    GTK_SIGNAL_FUNC(adj_adjust_destroyed),
				    (GtkObject*) skin_adj);
	if (adj != NULL && GTK_IS_ADJUSTMENT(adj))
		skin_adj->adj_connect_id
			= gtk_signal_connect_object
					   (GTK_OBJECT(adj),
					    "value-changed",
					    GTK_SIGNAL_FUNC(adj_value_changed),
					    (GtkObject*) skin_adj);
	else
		skin_adj->adj_connect_id = 0;
	return skin_adj;
}

static void
interpret_adj_option(const char* name, GtkObject* object, GdamSkin* skin,
                     GdamSkinAdjustment* skin_adj, gboolean* use_to_init)
{
	if (strcmp(name, "progress_bar_bottom_to_top") == 0) {
		gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(object),
		                                 GTK_PROGRESS_BOTTOM_TO_TOP);
		*use_to_init = FALSE;
		return;
	}
	if (strcmp(name, "progress_bar_top_to_bottom") == 0) {
		gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(object),
		                                 GTK_PROGRESS_TOP_TO_BOTTOM);
		*use_to_init = FALSE;
		return;
	}
	if (strcmp(name, "progress_bar_left_to_right") == 0) {
		gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(object),
		                                 GTK_PROGRESS_LEFT_TO_RIGHT);
		*use_to_init = FALSE;
		return;
	}
	if (strcmp(name, "progress_bar_right_to_left") == 0) {
		gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(object),
		                                 GTK_PROGRESS_RIGHT_TO_LEFT);
		*use_to_init = FALSE;
		return;
	}
	if (strcmp(name, "int") == 0) {
		skin_adj->is_int = TRUE;
		return;
	}
	if (strcmp(name, "inverted") == 0) {
		skin_adj->inverted = TRUE;
		return;
	}
	if (strcmp(name, "force") == 0) {
		*use_to_init = TRUE;
		return;
	}
}
	
	

                      

static gboolean gdam_skin_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	int i;
	GdamArg arg;
	GtkAdjustment* adjustment;
	gboolean use_to_initialize = FALSE;
	GdamSkinAdjustment* skin_adj;

	g_return_val_if_fail(handler_names[1] != NULL, TRUE);

	adjustment = util_get_widget_adjustment(GTK_WIDGET(object));
	g_return_val_if_fail(adjustment != NULL, TRUE);

	skin_adj = gdam_skin_adjustment_new(GTK_OBJECT(skin->model),
	                                    GTK_WIDGET(object),
					    handler_names[1],
					    FALSE);

	for (i = 2; i < num_handler_names; i++)
		interpret_adj_option(handler_names[i], object, skin, skin_adj,
		                     &use_to_initialize);

	if (!use_to_initialize) {
		const char* arg = handler_names[1];
		int num_out;
		GdamArg* args;
		gdam_arg_get_by_name(GTK_OBJECT(skin->model),
		                     1, &arg,
				     &num_out, &args);
		if (num_out == 1) {
			gdouble val = 0.0;
			if (args->type == GDAM_ARG_TYPE_DOUBLE) {
				val = GDAM_ARG_DOUBLE(args);
			} else if (args->type == GDAM_ARG_TYPE_INT) {
				val = GDAM_ARG_INT(args);
			} else {
				g_warning("type mismatch in gdamskin adj");
				return TRUE;
			}
			val = gdam_skin_adjustment_get_value(skin_adj, val);
			if (skin_adj->do_widget_change != NULL)
				(*skin_adj->do_widget_change)(skin_adj, val);
			else
				gtk_adjustment_set_value(skin_adj->adjustment, 
				                         val);
		}
	}

	if (use_to_initialize) {
		int type;
		float val = adjustment->value;
		arg.name = handler_names[1];
		type = gdam_arg_query_type(
				GTK_OBJECT(skin->model)->klass->type, 
			        arg.name);
		if (type == GDAM_ARG_TYPE_DOUBLE) {
			arg.type = GDAM_ARG_TYPE_DOUBLE;
			GDAM_ARG_DOUBLE(&arg) = val;
		}
		else if (type == GDAM_ARG_TYPE_INT) {
			arg.type = GDAM_ARG_TYPE_INT;
			GDAM_ARG_INT(&arg) = (int)val;
		}
		else {
			g_warning("type mismatch in gdamskin adj handler");
			return TRUE;
		}
		arg.name = handler_names[1];
		gdam_arg_set_object_arg(GTK_OBJECT(skin->model), &arg);
	}
	return TRUE;
}


/*
 *                  ``gdam_skin_label''
 *
 * Keep a label updated with a certain arguments
 * floating point or string value. 
 */
typedef struct _GdamSkinLabelSetter GdamSkinLabelSetter;
struct _GdamSkinLabelSetter {
	char*			arg_name;
	GtkObject*		object;
	GtkWidget*		label;
	char*			format;
	gboolean		int_format;
	gdouble			scale;
	guint			label_destroy_id;
	guint			object_destroy_id;
	guint			arg_connect_id;
};

static void free_skin_label_setter(GdamSkinLabelSetter* setter)
{
	if (setter->object != NULL) {
		gdam_arg_disconnect(setter->object, setter->arg_connect_id);
		gtk_signal_disconnect(setter->object,
		                      setter->object_destroy_id);
		setter->object = NULL;
	}
	if (setter->label != NULL) {
		gtk_signal_disconnect(GTK_OBJECT(setter->label),
		                      setter->label_destroy_id);
		setter->label = NULL;
	}
	g_free(setter->format);
	g_free(setter->arg_name);
	g_free(setter);
}
static void label_destroyed(GdamSkinLabelSetter* setter)
{
	setter->label = NULL;
	free_skin_label_setter(setter);
}

static void label_object_destroyed(GdamSkinLabelSetter* setter)
{
	setter->object = NULL;
	free_skin_label_setter(setter);
}
static void label_setter_set_text(GdamSkinLabelSetter* label_setter,
                                  const char*          txt)
{
	GtkWidget* widget = label_setter->label;
	if (txt == NULL)
		return;
	if (GTK_IS_LABEL(widget))
		gtk_label_set_text(GTK_LABEL(widget), txt);
	else if (GTK_IS_ENTRY(widget))
		gtk_entry_set_text(GTK_ENTRY(widget), txt);
	else {
		g_warning("widget was neither an entry nor a label");
		return;
	}
}

static void label_setter_arg_changed(gpointer setter, GdamArg* arg,
                                     GtkObject* object)
{
	GdamSkinLabelSetter* label_setter = setter;
	gdouble val;
	char* new_label;
	if (arg->type == GDAM_ARG_TYPE_STRING) {
		label_setter_set_text(label_setter, GDAM_ARG_STRING(arg));
		return;
	}
	val = GDAM_ARG_DOUBLE(arg) * label_setter->scale;
	if (label_setter->int_format) {
		new_label = g_strdup_printf(label_setter->format, (int)val);
	} else {
		new_label = g_strdup_printf(label_setter->format, val);
	}
	gdam_debug("label_setter: %s", new_label);
	label_setter_set_text(label_setter, new_label);
	g_free(new_label);
}

static gboolean gdam_skin_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamSkinLabelSetter* setter;
	setter = g_new(GdamSkinLabelSetter, 1);
	setter->arg_name = g_strdup(handler_names[1]);
	setter->object = GTK_OBJECT(skin->model);
	setter->label = GTK_WIDGET(object);
	setter->format = g_strdup(handler_names[2]);
	setter->scale = handler_names[3] ? atof(handler_names[3]) : 1.0;
	setter->int_format = num_handler_names >= 5 ? atoi(handler_names[4])
	                                            : FALSE;
	setter->label_destroy_id
		= gtk_signal_connect_object(object, "destroy",
				    GTK_SIGNAL_FUNC(label_destroyed),
					    (GtkObject*) setter);
	setter->object_destroy_id
		= gtk_signal_connect_object(setter->object,
		                            "destroy",
				    GTK_SIGNAL_FUNC(label_object_destroyed),
					    (GtkObject*) setter);
	setter->arg_connect_id
		= gdam_arg_connect(setter->object,
		                   setter->arg_name,
				   label_setter_arg_changed,
				   setter);
	{
		GdamArg* args;
		int num_args;
		gdam_arg_get_by_name(setter->object,
		                     1, (const char* const*)&setter->arg_name,
				     &num_args, &args);
		if (num_args == 1) {
			label_setter_arg_changed(setter, args, setter->object);
		} else {
			g_warning("no argument %s in %s (gdam_skin_label)",
				setter->arg_name,
				gtk_type_name(setter->object->klass->type));
		}
	}
	return TRUE;
}


/*
 *                   ``gdam_skin_toggle_button''
 *
 * For a toggling integer state that's attached to a GtkToggleButton.
 *
 *
 *                   ``gdam_skin_do_toggle''
 *
 * For button that toggles an argument.
 */
typedef struct _ToggledData ToggledData;
struct _ToggledData {
	GtkObject* object;
	GtkToggleButton* toggle;
	guint toggle_connect_id;
	guint arg_connect_id;
	guint toggle_destroy_id;
	guint arg_destroy_id;
	char* arg_name;
};
static void free_toggled_data(gpointer ptr)
{
	ToggledData* toggled = ptr;
	g_free(toggled->arg_name);
	if (toggled->arg_connect_id != 0) {
		gdam_arg_disconnect(toggled->object, toggled->arg_connect_id);
		gtk_signal_disconnect(toggled->object, toggled->arg_destroy_id);
	}
	if (toggled->toggle_connect_id != 0) {
		GtkObject* tog = GTK_OBJECT(toggled->toggle);
		gtk_signal_disconnect(tog, toggled->toggle_connect_id);
		gtk_signal_disconnect(tog, toggled->toggle_destroy_id);
	}
	g_free(ptr);
}

static void toggle_must_change(ToggledData* data, GdamArg* arg, GtkObject* obj)
{
	GtkObject* toggle = GTK_OBJECT(data->toggle);
	gtk_signal_handler_block(toggle, data->toggle_connect_id);
	gtk_toggle_button_set_active(data->toggle, GDAM_ARG_INT(arg));
	gtk_signal_handler_unblock(toggle, data->toggle_connect_id);
}

static void toggled(ToggledData* toggled)
{
	GdamArg arg;
	arg.name = toggled->arg_name;
	arg.type = GDAM_ARG_TYPE_INT;
	if (toggled->toggle)
		GDAM_ARG_INT(&arg) = toggled->toggle->active;
	else {
		GdamArg* cur_arg;
		int n_args;
		const char* name = (const char*)toggled->arg_name;
		gdam_arg_get_by_name(toggled->object,
				     1,
				     &name,
				     &n_args,
				     &cur_arg);
		g_return_if_fail(n_args == 1);
		GDAM_ARG_INT(&arg) = !GDAM_ARG_INT(cur_arg);
		gdam_arg_free_array(1,cur_arg);
	}
	gdam_debug("gdamskin: arg toggled: %s", GDAM_ARG_INT(&arg) ? 
							"true" : "false");
	gdam_arg_set_object_arg(toggled->object, &arg);
}

static void toggle_died(ToggledData* data)
{
	data->toggle_destroy_id = 0;
	data->toggle_connect_id = 0;
	free_toggled_data(data);
}
static void toggle_object_died(ToggledData* data)
{
	data->arg_destroy_id = 0;
	data->arg_connect_id = 0;
	free_toggled_data(data);
}
static gboolean gdam_skin_toggle_button_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	ToggledData* toggled_data = g_new(ToggledData, 1);
	toggled_data->object = GTK_OBJECT(skin->model);
	toggled_data->toggle = GTK_TOGGLE_BUTTON(object);
	toggled_data->arg_name = g_strdup(handler_names[1]);
	{
		GdamArg* args;
		int n_args;
		gdam_arg_get_by_name(toggled_data->object,
		                     1,
				     (const char*const*)handler_names + 1,
				     &n_args,
				     &args);
		if (n_args != 1) {
			g_warning("no %s in %s", handler_names[1],
			    gtk_type_name(toggled_data->object->klass->type));
			return TRUE;
		}
		gtk_toggle_button_set_active(toggled_data->toggle,
		                             GDAM_ARG_INT(args));
		gdam_arg_free_array(1, args);
	}
	toggled_data->toggle_connect_id
			= gtk_signal_connect_full(object,
				                  signal_name,
				                  GTK_SIGNAL_FUNC(toggled),
				                  NULL,
				                  toggled_data,
				                  NULL,
				                  TRUE,
				                  FALSE);
	toggled_data->arg_connect_id
			= gdam_arg_connect(toggled_data->object,
	                                   handler_names[1],
			                   GDAM_ARG_FUNC(toggle_must_change),
			                   toggled_data);
	toggled_data->toggle_destroy_id
			= gtk_signal_connect_object
				 (object,
	                          "destroy",
			          GTK_SIGNAL_FUNC(toggle_died),
			          (GtkObject*) toggled_data);
	toggled_data->arg_destroy_id
			= gtk_signal_connect_object
				 (toggled_data->object,
	                          "destroy",
			          GTK_SIGNAL_FUNC(toggle_object_died),
			          (GtkObject*) toggled_data);
	return TRUE;
}

static gboolean gdam_skin_do_toggle_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	ToggledData* toggled_data = g_new0(ToggledData, 1);
	toggled_data->object = GTK_OBJECT(skin->model);
	toggled_data->toggle = NULL;
	toggled_data->arg_name = g_strdup(handler_names[1]);
	gtk_signal_connect_full(object,
				signal_name,
				GTK_SIGNAL_FUNC(toggled),
				NULL,
				toggled_data,
				free_toggled_data,
				TRUE,
				FALSE);
	return TRUE;
}


/*
 *                       ``GdamSkinHandler''
 *
 *                              ====>  gdam_skin_arg_double
 *                              ====>  gdam_skin_arg_string
 *
 * Encapsulate constant-argument based handlers.
 *
 * For example a button which sets an argument to a constant value
 * that is specified in the xml.
 */

typedef struct _GdamSkinArgHandler GdamSkinArgHandler;
struct _GdamSkinArgHandler {
	GdamArg*		to_set;
	GtkObject*		object;
};

static void free_skin_arg_handler(gpointer ptr)
{
	GdamSkinArgHandler* handler = ptr;
	gdam_arg_free_array(1, handler->to_set);
	g_free(handler);
}

static void on_arg_event(GdamSkinArgHandler* handler)
{
	gdam_arg_set_object_arg(handler->object, handler->to_set);
}
	
static gboolean gdam_skin_arg_string_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamArg* arg;
	GdamSkinArgHandler* handler;
	if (num_handler_names < 3) {
		g_warning("usage: gdam_skin_arg_string:param:value");
		return TRUE;
	}
	arg = g_new(GdamArg, 1);

	arg->type = GDAM_ARG_TYPE_STRING;
	arg->name = g_strdup(handler_names[1]);
	GDAM_ARG_STRING(arg) = g_strdup(handler_names[2]);
	handler = g_new(GdamSkinArgHandler, 1);
	handler->to_set = arg;
	handler->object = GTK_OBJECT(skin->model);
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(on_arg_event),
				NULL,
				handler,
				free_skin_arg_handler,
				TRUE,
				FALSE);
	return TRUE;
}

static gboolean gdam_skin_arg_int_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamArg* arg;
	GdamSkinArgHandler* handler;
	char* endp;
	gdouble val;
	if (num_handler_names < 3) {
		g_warning("usage: gdam_skin_arg_double:param:value");
		return TRUE;
	}
	arg = g_new(GdamArg, 1);

	val = strtod(handler_names[2], &endp);
	if (*endp != '\0') {
		g_warning("error parsing value %s (for %s:%s)",
		                                         handler_names[2],
		                                         handler_names[0],
		                                         handler_names[1]);
	}
	arg->type = GDAM_ARG_TYPE_INT;
	arg->name = g_strdup(handler_names[1]);
	GDAM_ARG_INT(arg) = (int)val;

	handler = g_new(GdamSkinArgHandler, 1);
	handler->to_set = arg;
	handler->object = GTK_OBJECT(skin->model);
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(on_arg_event),
				NULL,
				handler,
				free_skin_arg_handler,
				TRUE,
				FALSE);
	return TRUE;
}
static void on_set_buffer_event(GdamSkinArgHandler* handler)
				
{
	GdamArg* arg = handler->to_set;
	GDAM_ARG_INT(arg) = gdam_buffer_get_default();
	gdam_arg_set_object_arg(handler->object, arg);
}

static gboolean gdam_skin_set_to_buffer_id_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamArg* arg;
	GdamSkinArgHandler* handler;
	if (num_handler_names < 2) {
		g_warning("usage: gdam_skin_set_to_buffer_id:param");
		return TRUE;
	}
	arg = g_new(GdamArg, 1);
	arg->type = GDAM_ARG_TYPE_INT;
	arg->name = g_strdup(handler_names[1]);
	handler = g_new(GdamSkinArgHandler, 1);
	handler->to_set = arg;
	handler->object = GTK_OBJECT(skin->model);
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(on_set_buffer_event),
				NULL,
				handler,
				free_skin_arg_handler,
				TRUE,
				FALSE);
	return TRUE;
}

static gboolean gdam_skin_arg_double_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamArg* arg;
	GdamSkinArgHandler* handler;
	char* endp;
	gdouble val;
	if (num_handler_names < 3) {
		g_warning("usage: gdam_skin_arg_double:param:value");
		return TRUE;
	}
	arg = g_new(GdamArg, 1);

	val = strtod(handler_names[2], &endp);
	if (*endp != '\0') {
		g_warning("error parsing value %s (for %s:%s)",
		                                         handler_names[2],
		                                         handler_names[0],
		                                         handler_names[1]);
	}
	arg->type = GDAM_ARG_TYPE_DOUBLE;
	arg->name = g_strdup(handler_names[1]);
	GDAM_ARG_DOUBLE(arg) = val;

	handler = g_new(GdamSkinArgHandler, 1);
	handler->to_set = arg;
	handler->object = GTK_OBJECT(skin->model);
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(on_arg_event),
				NULL,
				handler,
				free_skin_arg_handler,
				TRUE,
				FALSE);
	return TRUE;
}

/*
 *                    gdam_skin_try_or_destroy
 *
 * This prefixes another command, ie
 *               gdam_skin_try_or_destroy:gdam_midi_patch:...
 *
 * It causes the widget it is attached to to be destroyed
 * if the handler that follows is unknown.
 *
 * This causes the example to display midi buttons only if
 * you have midi.
 *
 * Used to flag widgets which should be conditionally constructed.
 *
 *
 *                     gdam_skin_destroy_if_available
 *
 * Does the opposite, useful for a widget that should look
 * one way, say w/ midi, and another way w/o.
 */
static gboolean gdam_skin_try_or_destroy_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkType skin_type = GTK_OBJECT(skin)->klass->type;
	if (num_handler_names < 2) {
		g_warning("%s needs 1 argument", handler_names[0]);
		return TRUE;
	}
	if (has_connect_handler(handler_names[1], skin_type)) {
		char* handler_joined = g_strjoinv(":", handler_names + 1);
		gboolean rv;
		rv = skin_generic_connect(handler_joined,
		                          object,
					  signal_name,
					  skin);
		if (!rv) {
			g_warning("error doing connect %s", handler_joined);
		}
		g_free(handler_joined);
	} else {
		gtk_object_destroy(object);
	}
	return TRUE;
}

static gboolean gdam_skin_destroy_if_available_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkType skin_type = GTK_OBJECT(skin)->klass->type;
	if (num_handler_names != 2) {
		g_warning("%s needs 1 argument", handler_names[0]);
		return TRUE;
	}
	if (has_connect_handler(handler_names[1], skin_type)) {
		gtk_object_destroy(object);
	}
	return TRUE;
}

static void dump_xml(GdamSkin* skin)
{
	xmlDoc* doc;
	GdamXmlInfo* info;
	xmlNode* node;
	xmlNode* top;
	info = gdam_xml_info_new();
	node = gdam_skin_to_xml(skin, info);
	gdam_xml_info_destroy(info);
	top = xmlNewNode(NULL, "GDAM-Skin-Interface");
	xmlAddChild(top, node);
	doc = xmlNewDoc("1.0");
	doc->root = top;
	g_return_if_fail(doc != NULL);
	xmlDocDump(stderr, doc);
	xmlSaveFile("tmp.xml", doc);
	xmlFreeDoc(doc);
}

static gboolean gdam_skin_write_xml_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object_while_alive
			(object,
			 signal_name,
			 GTK_SIGNAL_FUNC(dump_xml),
			 GTK_OBJECT(skin));
	return TRUE;
}


static GdamSkinGenericConnect connect_infos[] =
{ 	{ "gdam_skin_toplevel",
	   gdam_skin_toplevel_handler },  
  	{ "gdam_skin_launch",
	   gdam_skin_launch_handler },  
  	{ "gdam_skin_launch_unique",
	   gdam_skin_launch_unique_handler },  
	{ "gdam_skin_destroy",
	   gdam_skin_destroy_handler },
	{ "gdam_skin_destroy_top",
	   gdam_skin_destroy_top_handler },
	{ "gdam_skin_raise_or_launch",
	   gdam_skin_raise_or_launch_handler },
	{ "gdam_skin_spin_button_display_beatstyle",
	   gdam_skin_spin_button_display_beatstyle_handler },
	{ "gdam_skin_adjustment", 
	   gdam_skin_adjustment_handler },
	{ "gdam_skin_arg_string",
	   gdam_skin_arg_string_handler },
	{ "gdam_skin_arg_double",
	   gdam_skin_arg_double_handler },
        { "gdam_skin_set_to_buffer_id",
           gdam_skin_set_to_buffer_id_handler },
	{ "gdam_skin_arg_int",
	   gdam_skin_arg_int_handler },
	{ "gdam_skin_write_xml",
	   gdam_skin_write_xml_handler },
	{ "gdam_skin_toggle_button",
	   gdam_skin_toggle_button_handler },
	{ "gdam_skin_do_toggle",
	   gdam_skin_do_toggle_handler },
	{ "gdam_skin_label",
	   gdam_skin_label_handler },
	{ "gdam_skin_try_or_destroy",
	   gdam_skin_try_or_destroy_handler },
	{ "gdam_skin_destroy_if_available",
           gdam_skin_destroy_if_available_handler },
	   							};

static void gdam_skin_class_init(GdamSkinClass* skin_class)
{
	GtkObjectClass* object_class = (GtkObjectClass*)skin_class;
	GdamViewClass* view_class;
	view_class = GDAM_VIEW_CLASS(skin_class);
	view_class->make_xml = gdam_skin_make_xml;
	view_class->parse_xml = gdam_skin_parse_xml;
	view_class->attach_to_model = gdam_skin_attach_to_model;
	object_class->destroy = gdam_skin_destroy;
	skin_class->initialize_main_widget = gdam_skin_initialize_main_widget;
	skin_class->build_new_model = gdam_skin_build_new_model;
	skin_class->model_type = GTK_TYPE_NONE;
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
}

static void gdam_skin_init(GdamSkinClass* skin_class)
{
}

GtkType gdam_skin_get_type()
{
	static GtkType skin_type = 0;
	if (!skin_type) {
		static const GtkTypeInfo skin_info = {
			"GdamSkin",
			sizeof(GdamSkin),
			sizeof(GdamSkinClass),
			(GtkClassInitFunc) gdam_skin_class_init,
			(GtkObjectInitFunc) gdam_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_VIEW);
		skin_type = gtk_type_unique(parent_class->type,
		                            &skin_info);
	}
	return skin_type;
}


static void gdam_skin_destroy(GtkObject* object)
{
	GdamSkin* skin = GDAM_SKIN(object);
	gdam_debug("gdam_skin_destroy: %s",
		gtk_type_name(object->klass->type));
	if (skin->model) {
		GtkObject* model = GTK_OBJECT(skin->model);
		skin->model = NULL;
		gtk_object_unref(model);
	}
	if (skin->main_widget) {
		gtk_object_destroy(GTK_OBJECT(skin->main_widget));
	}
	(*parent_class->destroy)(object);
}
/*
 *                     XML reading mayhem.
 *
 * Functions to:
 *       - find xml files.
 *       - wrap libglade's autoconnect mechanism
 */


static GHashTable* xml_file_paths = NULL;

const char* gdam_skin_search_for_file(const char* filename)
{
	char* path;
	if (xml_file_paths == NULL)
		xml_file_paths = g_hash_table_new(g_str_hash, g_str_equal);
	path = g_hash_table_lookup(xml_file_paths, filename);
	if (path != NULL)
		return path;
	path = gdam_search_path(filename,
				xml_path,
			 GDAM_SEARCH_PATH_FIND_FILE);
	if (path != NULL)
		g_hash_table_insert(xml_file_paths, g_strdup(filename), path);
	return path;
}

/*
 * Autodetect the type of xml file, and open appropriately.
 */
gboolean gdam_skin_open(const char*  filename,
                        GdamChannel* channel)
{
	GdamModel* rv_model;
	xmlDoc* doc;
	const char* main_name;
	doc = xmlParseFile(filename);
	g_return_val_if_fail(doc != NULL, FALSE);
	g_return_val_if_fail(doc->root != NULL, FALSE);
	main_name = doc->root->name;

	if (strcmp(main_name, "GTK-Interface") == 0) {
		if (gdam_skin_new(GDAM_TYPE_SKIN,
				  filename,
				  0, NULL,
				  channel,
				  &rv_model) == NULL)
			return FALSE;
		return TRUE;
	}

	if (strcmp(main_name, "GDAM-Component-Interface") == 0) {
		GdamModel* model;
		GSList* skins;
		model = gdam_model_from_xml(doc->root->childs, channel, NULL);
		for (skins = model->skin_list;
		     skins != NULL;
		     skins = skins->next) {
		      	if (GDAM_IS_SKIN(skins->data))
				launch_widget(GDAM_SKIN(skins->data));
		}
		gdam_toplevel_add_child(gdam_toplevel_get_global(channel),
		                        model);
		return TRUE;
	}

	if (strcmp(main_name, "GDAM-Skin-Interface") == 0) {
		GdamSkin* skin;
		skin = gdam_skin_from_xml(doc->root->childs, channel, NULL);
		launch_widget(skin);
		return TRUE;
	}

	return TRUE;
}

static void null_main_widget_and_unref_skin(GdamSkin* skin)
{
	GtkWidget* main_widget = skin->main_widget;
	g_return_if_fail (main_widget != NULL);
	skin->main_widget = NULL;
	gtk_object_unref(GTK_OBJECT(skin));
}

/* This is the format of the `user_data' parameter to gdam_skin_connect_func */
typedef struct _SkinConnectData SkinConnectData;
struct _SkinConnectData {
	GSList*			objects_refed;
	GdamSkin*		skin;
};

void            gdam_skin_connect_func  (const gchar*     handler,
                                         GtkObject*       object,
                                         const gchar*     signal_name,
				         const gchar*     signal_data,
                                         GtkObject*       connect_object,
				         gboolean         after,
                                         gpointer         user_data)
{
	GdamSkin* skin;
	SkinConnectData* connect_data = user_data;
	skin = GDAM_SKIN(connect_data->skin);
	g_return_if_fail(skin != NULL);
	if (GTK_OBJECT_DESTROYED(object))
		return;
	connect_data->objects_refed
		= g_slist_prepend(connect_data->objects_refed, object);
	gtk_object_ref(object);
	if (skin_generic_connect(handler, object, signal_name, skin))
		return;

	g_warning("no handler for `%s' in %s\n\t[signal=%s, object type=%s]",
		handler,
		gtk_type_name(GTK_OBJECT(user_data)->klass->type),
		signal_name,
		gtk_type_name(object->klass->type));
}


static gboolean gdam_skin_initialize_main_widget
				(GdamSkin*    skin,
				 const char*  skin_filename)
{
	const char* skin_fname;
	GladeXML* xml;
	SkinConnectData data;
	/*
	 * Read the XML file, build the widgetry.
	 */
	skin_fname = gdam_skin_search_for_file(skin_filename);

	if (skin_fname == NULL){
		g_warning("file %s wasn't found in your path", skin_filename);
		return FALSE;
	}

	if (!gdam_file_exists(skin_fname)) {
		g_warning("skin %s doesn't exist", skin_fname);
		return FALSE;
	}

    	xml = glade_xml_new(skin_fname, NULL);
	if (xml == NULL) {
		g_warning("couldn't create widgets from %s", skin_fname);
		return FALSE;
	}

	data.skin = skin;
	data.objects_refed = NULL;
    	glade_xml_signal_autoconnect_full(xml, gdam_skin_connect_func, &data);
	g_slist_foreach(data.objects_refed, (GFunc) gtk_object_unref, NULL);
	g_slist_free(data.objects_refed);


	/*
	 * Find the main_widget.
	 */
    	skin->main_widget = glade_xml_get_widget(xml, "main_widget");
    	if (!skin->main_widget) {
		g_warning("couldn't get main_widget");
		return FALSE;
	}

	return TRUE;
}


/*
 * Reading/writing of GDAM xml trees...
 */
void            gdam_skin_add_filename_xml (GdamSkin*        skin,
					    xmlNode*         parent)
{
	g_return_if_fail(skin->filename != NULL);
	xmlNewChild(parent, NULL, "filename", skin->filename);
		
}
void            gdam_skin_handle_filename_xml(GdamSkin*        skin,
					      xmlNode*         node)
{
	const char* skin_fname = NULL;
	if (node->childs != NULL)
		skin_fname = node->childs->content;
	if (skin_fname == NULL) {
		g_warning("filename not a text node: "
				 "malformed");
		return;
	}
	skin->filename = g_strdup(skin_fname);
	if (!gdam_skin_initialize_widgets(skin)) {
		g_warning("error initing widgets");
		return;
	}
	if (skin->main_widget == NULL)
		g_warning("initialize widgets left main_widget NULL");
}

static gboolean gdam_skin_make_xml(GdamView*               view,
                                   xmlNode*                parent,
				   GdamXmlInfo*            info)
{
	GdamSkin* skin = GDAM_SKIN(view);
	if (skin->model != NULL
	 && gdam_xml_info_lookup_by_object(info, skin->model) == 0) {
		xmlNode* node;
		xmlNode* comp_node;
		node = xmlNewNode(NULL, "model");
		comp_node = gdam_model_to_xml(skin->model, info);
		xmlAddChild(node, comp_node);
		xmlAddChild(parent, node);
	}
	gdam_skin_add_filename_xml(skin, parent);
	return TRUE;
}

static gboolean gdam_skin_parse_xml(GdamView*               view,
                                    xmlNode*                children,
				    GdamXmlInfo*            xml_info)
{
	xmlNode* child;
	GtkObject* object = GTK_OBJECT(view);
	GdamViewClass* view_class;
	GdamSkin* skin = GDAM_SKIN(view);
	view_class = GDAM_VIEW_CLASS(object->klass);
	for (child = children; child != NULL; child = child->next) {
		if (strcmp(child->name, "model") == 0) {
			GdamModel* model;
			if (skin->model != NULL) {
				g_message("only one model per skin");
				continue;
			}
			model = gdam_model_from_xml(child->childs, 
					            skin->channel,
						    xml_info);
			if (view_class->attach_to_model != NULL)
				(*view_class->attach_to_model)(view, model);
			gdam_view_add(GDAM_VIEW(skin), model);
			continue;
		}
		if (strcmp(child->name, "filename") == 0) {
			gdam_skin_handle_filename_xml(skin, child);
		}
	}
	return TRUE;
}

xmlNode* gdam_skin_to_xml(GdamSkin* skin, GdamXmlInfo* info)
{
	xmlNode* rv;
	rv = gdam_view_to_xml(GDAM_VIEW(skin), info);
	g_return_val_if_fail(rv != NULL, NULL);
	return rv;
}
GdamSkin*      gdam_skin_from_xml(xmlNode*      node,
                                  GdamChannel*  channel,
				  GdamXmlInfo*  xml_info)
{
	GdamSkin* skin;
	GtkType type;
	if (strcmp(node->name, "id") == 0) {
		const char* content;
		int id;
		if (node->content != NULL)
			content = node->content;
		else if (node->childs && node->childs->content)
			content = node->childs->content;
		else
			return NULL;
		id = atoi(content);
		return GDAM_SKIN(gdam_xml_info_lookup_by_id(xml_info, id));
	}
	type = gdam_plugin_get_type_by_name("skin", node->name);
	if (type == 0) {
		g_warning("no skintype named %s could be found", node->name);
		return NULL;
	}
	skin = GDAM_SKIN(gtk_type_new(type));
	skin->channel = channel;
	if (!gdam_view_parse_xml(GDAM_VIEW(skin), node->childs, xml_info)) {
		g_warning("error parsing xml");
		gtk_object_sink(GTK_OBJECT(skin));
		return NULL;
	}
	return skin;
}
static gboolean gdam_skin_initialize_widgets(GdamSkin* skin)
{
	GtkObject* object = GTK_OBJECT(skin);
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(object->klass);
	gtk_object_ref((GtkObject*)skin);
	/* This is where the XML gets read for xml-based skins. */
	if (! (*skin_class->initialize_main_widget)(skin, skin->filename) ) {
		gtk_object_unref((GtkObject*)skin);
		return FALSE;
	}

	gtk_object_sink((GtkObject*)skin);
	gtk_signal_connect_object(GTK_OBJECT(skin->main_widget),
	                          "destroy",
			  GTK_SIGNAL_FUNC(null_main_widget_and_unref_skin),
				  (GtkObject*) skin);
	gtk_signal_connect_object_while_alive
	                         (GTK_OBJECT(skin),
	                          "destroy",
				  GTK_SIGNAL_FUNC(gtk_object_destroy),
				  GTK_OBJECT(skin->main_widget));

	/*
	 * Invoke the "finished_connecting" handler;
	 * indicate we're done.
	 */
	skin->finished_connecting = 1;
	if (skin_class->post_connecting_handler)
		(*skin_class->post_connecting_handler)(skin);
	gdam_skin_run_pending_initializers(skin);
	return TRUE;
}

/*=====================================================================*
 *                            gdam_skin_new                            *
 *                                                                     *
 * Parameters:                                                         *
 *     skin_type         the type of skin to instantiate.              *
 *=====================================================================*/
GdamSkin*       gdam_skin_new           (GtkType          skin_type,
                                         const char*      skin_filename,
					 int              num_options,
					 char**           options,
					 GdamChannel*     channel,
					 GdamModel**  model_out)
{
	GdamSkin* skin = GDAM_SKIN(gtk_type_new(skin_type));
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(gtk_type_class(skin_type));

	skin->channel = channel;

	/*
	 * Build the skin's underlying model, if it has one.
	 */
	if (skin_class->build_new_model) {
		if (! (*skin_class->build_new_model)(skin, channel,
		                                         num_options,
							 options)) {
			g_warning("error building model");
			gtk_object_sink(GTK_OBJECT(skin));
			return NULL;
		}
		if (skin->model) {
			gtk_object_ref((GtkObject*)(skin->model));
			gtk_object_sink((GtkObject*)(skin->model));
			gdam_view_add(GDAM_VIEW(skin), skin->model);
		}
	} else
		skin->model = NULL;
	*model_out = skin->model;

	skin->filename = g_strdup(skin_filename);
	if (!gdam_skin_initialize_widgets(skin)) {
		gtk_object_sink((GtkObject*)skin);
		gtk_object_unref((GtkObject*)skin);
		return NULL;
	}

	return skin;
}

GdamSkin* gdam_skin_new_by_type_name(const char*      skin_type_name,
                                     const char*      skin_filename,
				     int              num_options,
				     char**           options,
				     GdamChannel*     channel,
			             GdamModel**  model_out)
{
	GtkType skin_type;
	skin_type = gdam_plugin_get_type_by_name("skin", skin_type_name);
	if (skin_type == 0) {
		g_warning("error getting skin for %s", skin_type_name);
		return NULL;
	}
	return gdam_skin_new(skin_type,
	                     skin_filename,
			     num_options,
			     options,
			     channel,
			     model_out);
}

/*
 *   ____                      _            _    _             _
 *  / ___| ___ _ __   ___ _ __(_) ___   ___| | _(_)_ __  _ __ (_)_ __   __ _
 * | |  _ / _ \ '_ \ / _ \ '__| |/ __| / __| |/ / | '_ \| '_ \| | '_ \ / _` |
 * | |_| |  __/ | | |  __/ |  | | (__  \__ \   <| | | | | | | | | | | | (_| |
 *  \____|\___|_| |_|\___|_|  |_|\___| |___/_|\_\_|_| |_|_| |_|_|_| |_|\__, |
 *                                                                     |___/
 */

static GHashTable*  type_info = NULL;

typedef struct _SkinTypeInfo SkinTypeInfo;
struct _SkinTypeInfo {
	/* Maps from names to GdamSkinSimpleConnect. */
	GHashTable*		simple_connect_infos;

	/* Maps from names to GdamSkinGenericConnect. */
	GHashTable*		generic_connect_infos;
};

static inline GHashTable* str_hash_table()
{
	return g_hash_table_new(g_str_hash, g_str_equal);
}

static SkinTypeInfo* get_skin_info(GtkType type)
{
	SkinTypeInfo* info;
	if (type_info) {
		info = g_hash_table_lookup(type_info, GINT_TO_POINTER(type));
	} else {
		info = NULL;
		type_info = g_hash_table_new(NULL, NULL);
	}
	if (info == NULL) {
		info = g_new0(SkinTypeInfo, 1);
		info->simple_connect_infos = str_hash_table();
		info->generic_connect_infos = str_hash_table();
		g_hash_table_insert(type_info, GINT_TO_POINTER(type), info);
	}
	return info;
}

void gdam_skin_class_register_simple_connections
                                 (GdamSkinClass*        skin_class,
				  int                   num_types,
				  GdamSkinSimpleConnect* connect_types)
{
	SkinTypeInfo* info;
	GtkType type = GTK_OBJECT_CLASS(skin_class)->type;
	int i;
	info = get_skin_info(type);
	for (i = 0; i < num_types; i++) {
		g_hash_table_insert(info->simple_connect_infos,
		                    (char*) connect_types[i].name,
				    &connect_types[i]);
		gdam_debug("registering %s in %s", connect_types[i].name,
				gtk_type_name(type));
	}
}

typedef struct _DeferredConnectData DeferredConnectData;
struct _DeferredConnectData {
	GtkObject*		object;
	char*			signal_name;
	int			num_commands;
	char**			commands;
};

static gboolean deferred_connect_data_marshal(GdamSkin*            skin,
                                              GdamSkinInitializer* initializer)
{
	DeferredConnectData* connect_data = initializer->user_data;
	GdamSkinConnectHandler handler;
	handler = (GdamSkinConnectHandler) (initializer->init_func);
	return (*handler)(skin,
	                  connect_data->object,
                          connect_data->signal_name,
                          connect_data->num_commands,
		          connect_data->commands);
}
static void deferred_connect_data_destroy(gpointer user_data)
{
	DeferredConnectData* connect_data = user_data;
	int i, num;
	char** commands;
	g_free(connect_data->signal_name);
	num = connect_data->num_commands;
	commands = connect_data->commands;
	for (i = 0; i < num; i++)
		g_free(commands[i]);
	g_free(commands);
	g_free(connect_data);
}

void gdam_skin_defer_connect_func(GdamSkin*              skin,
                                  GtkObject*             object,
                                  GdamSkinConnectHandler handler,
			          const char*            signal_name,
			          int                    num_commands,
			          char**                 commands)
{
	GdamSkinInitializer* initializer;
	DeferredConnectData* connect_data;
	connect_data = g_new(DeferredConnectData, 1);
	connect_data->object = object;
	connect_data->signal_name = g_strdup(signal_name);
	connect_data->num_commands = num_commands;
	connect_data->commands = commands;

	initializer = g_new(GdamSkinInitializer, 1);
	initializer->init_func = (GdamSkinInitFunc) handler;
	initializer->destroyer = deferred_connect_data_destroy;
	initializer->marshal    = deferred_connect_data_marshal;
	initializer->user_data = connect_data;

	skin->initializer_list =
		g_slist_prepend(skin->initializer_list, initializer);
}

void gdam_skin_class_register_generic_handlers
                                 (GdamSkinClass*        skin_class,
				  int                   num_types,
				  GdamSkinGenericConnect* connect_types)
{
	SkinTypeInfo* info;
	int i;
	info = get_skin_info(GTK_OBJECT_CLASS(skin_class)->type);
	for (i = 0; i < num_types; i++)
		g_hash_table_insert(info->generic_connect_infos,
		                    (char*) connect_types[i].name,
				    &connect_types[i]);
}

static gboolean has_one_type    (GtkType         type,
                                 const gchar*    handler_name)
{
	SkinTypeInfo* skin_info;
	skin_info = g_hash_table_lookup(type_info, GINT_TO_POINTER(type));
	if (skin_info == NULL) {
		return FALSE;
	}
	if (g_hash_table_lookup(skin_info->generic_connect_infos,
	                        handler_name))
		return TRUE;
	if (g_hash_table_lookup(skin_info->simple_connect_infos,
	                        handler_name))
		return TRUE;
	return FALSE;
}

static gboolean connect_one_type(GtkType         type,
                                 const gchar*    handler,
                                 GtkObject*      object,
				 const gchar*    signal_name,
				 GdamSkin*       skin)
{
	GdamSkinSimpleConnect* simple_info;
	SkinTypeInfo* skin_info;
	skin_info = g_hash_table_lookup(type_info, GINT_TO_POINTER(type));
	if (skin_info == NULL) {
		/*gdam_debug("no type info for %s", gtk_type_name(type));*/
		return FALSE;
	}

	/*gdam_debug("trying %s in %s", handler, gtk_type_name(type));*/

	/*
	 * Try generic connections.
	 */
	{
		char** commands = g_strsplit(handler, ":", 0);
		GdamSkinGenericConnect* info;
		int num_commands = 0, i;
		info = g_hash_table_lookup(skin_info->generic_connect_infos,
					   commands[0]);
		while (commands[num_commands] != NULL)
			num_commands++;
		if (info) {
			if (!(*info->skin_func)(skin, object, signal_name,
			                 num_commands, commands)) {
				gdam_skin_defer_connect_func(skin,
				                             object,
							     info->skin_func,
				                             signal_name,
							     num_commands,
							     commands);
				return TRUE;
			}
			for (i = 0; i < num_commands; i++)
				g_free(commands[i]);
			g_free(commands);
			return TRUE;
		}
		for (i = 0; i < num_commands; i++)
			g_free(commands[i]);
		g_free(commands);
	}

	/*
	 * Try simple connection.
	 */
	simple_info = g_hash_table_lookup(skin_info->simple_connect_infos,
	                                  handler);
	if (simple_info) {
		gtk_signal_connect_object(object,
		                          signal_name,
				  GTK_SIGNAL_FUNC(simple_info->skin_func),
					  GTK_OBJECT(skin));
		return TRUE;
	}

	return FALSE;
}

static gboolean has_connect_handler(const gchar*    handler_name,
                                    GtkType         type)
{
	if (type_info == NULL)
		return FALSE;
	for (      ; type != 0; type = gtk_type_parent(type)) {
		if (has_one_type(type, handler_name))
			return TRUE;
	}
	return FALSE;
}

static gboolean skin_generic_connect(const gchar*    handler,
                                      GtkObject*      object,
				      const gchar*    signal_name,
				      GdamSkin*       skin)
{
	GtkType type;
	if (type_info == NULL)
		return FALSE;

	/* This triggers certain initialization, e.g. adding
	 * a `right-click' signal to button. */
	signal_name_hacks(signal_name, object);

	for (type = GTK_OBJECT(skin)->klass->type;
	      type != 0;
	      type = gtk_type_parent(type)) {
		if (connect_one_type(type, handler, object, signal_name, skin))
			return TRUE;
	}
	return FALSE;
}

void gdam_skin_connect_toplevel_notify (GdamSkin*             skin,
                                        GdamToplevel*         toplevel)
{
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(GTK_OBJECT(skin)->klass);
	if (skin_class->toplevel_connect)
		(*skin_class->toplevel_connect)(skin, toplevel);
	gdam_skin_run_pending_initializers(skin);
}

gboolean gdam_skin_do_connect    (GdamSkin*        skin,
                                  GtkObject*       object,
			          const char*      signal_name,
			          int              num_handler_names,
			          char**           handler_names)
{
	char* join = g_strjoinv(":", handler_names);
	gboolean rv = skin_generic_connect(join, object, signal_name, skin);
	g_free(join);
	return rv;
}

void gdam_skin_add_initializer(GdamSkin*       skin,
			       GdamSkinInitFunc initializer,
			       GDestroyNotify   destructor,
			       gpointer         user_data)
{
	GdamSkinInitializer* init = g_new(GdamSkinInitializer, 1);
	init->init_func  = initializer;
	init->marshal    = NULL;
	init->destroyer  = destructor;
	init->user_data  = user_data;
	skin->initializer_list = g_slist_prepend(skin->initializer_list, init);
}
static inline gboolean gdam_skin_call_init_func(GdamSkin*            skin,
                                                GdamSkinInitializer* init)
{
	if (init->marshal    != NULL)
		return (*init->marshal)(skin, init);
	else 
		return (*init->init_func)(skin, init->user_data);
}

void gdam_skin_run_pending_initializers(GdamSkin* skin)
{
	gboolean had_success;
	do {
		GSList* tried = NULL;
		had_success = FALSE;
		while (skin->initializer_list != NULL) {
			GSList* at = skin->initializer_list;
			GdamSkinInitializer* init = at->data;
			skin->initializer_list = at->next;
			if (gdam_skin_call_init_func(skin, init)) {
				if (init->destroyer != NULL)
					(*init->destroyer)(init->user_data);
				g_slist_free_1(at);
				g_free(init);
				had_success = TRUE;
			} else {
				at->next = tried;
				tried = at;
			}
		}
		skin->initializer_list = tried;
	} while (had_success);
	if (skin->initializer_list != NULL)
		gdam_debug("run deferred commands:  commands were pending");
}

GdamSkin* gdam_skin_find_from_model(GdamModel* model)
{
	GSList* list;
	g_return_val_if_fail(model != NULL, NULL);
	for (list = model->skin_list; list != NULL; list = list->next) {
		if (GDAM_IS_SKIN(list->data))
			return GDAM_SKIN(list->data);
	}
	return NULL;
}

typedef struct _Emitter Emitter;
struct _Emitter {
	GtkObject* object;
	guint signal_id;
};
static gboolean do_emit(gpointer emitter_ptr)
{
	Emitter* emitter = emitter_ptr;
	gtk_signal_emit(emitter->object, emitter->signal_id);
	gtk_object_unref(emitter->object);
	g_free(emitter);
	return FALSE;
}
	
static void in_idle_signal_emit(GtkObject* object, guint signal_id)
{
	Emitter* emitter = g_new(Emitter, 1);
	gtk_object_ref(object);
	emitter->object = object;
	emitter->signal_id = signal_id;
	g_idle_add(do_emit, emitter);
}


/* Simulated (ie left/right/middle button) clicks. */

static guint left_clicked_id = 0;
static guint middle_clicked_id = 0;
static guint right_clicked_id = 0;
static guint control_clicked_id = 0;


static gint map_event_to_gdam_clicked(GtkObject* object,
                                       GdkEventButton* event)
{
	if ((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK) {
		in_idle_signal_emit(object, control_clicked_id);
		return 1;
	}
	if (event->button == 1) {
		in_idle_signal_emit(object, left_clicked_id);
		return 1;
	}
	if (event->button == 2) {
		in_idle_signal_emit(object, middle_clicked_id);
		return 1;
	}
	if (event->button == 3) {
		in_idle_signal_emit(object, right_clicked_id);
		return 1;
	}
	return 0;
}

GHashTable* sim_clicked_widgets = NULL;

static void sim_clicked_destroyed(GtkWidget* widget)
{
	g_hash_table_remove(sim_clicked_widgets, widget);
}

static void attach_simulated_clicks(GtkWidget* widget)
{
	g_return_if_fail(widget != NULL);

	if (GTK_OBJECT_DESTROYED(widget))
		return;

	if (sim_clicked_widgets == NULL)
		sim_clicked_widgets = g_hash_table_new(NULL, NULL);
	if (g_hash_table_lookup(sim_clicked_widgets, widget))
		return;
	g_hash_table_insert(sim_clicked_widgets, widget, widget);
	gtk_signal_connect(GTK_OBJECT(widget),
			   "destroy",
	                   GTK_SIGNAL_FUNC(sim_clicked_destroyed),
			   NULL);
	gtk_signal_connect(GTK_OBJECT(widget),
			   "button-press-event",
	                   GTK_SIGNAL_FUNC(map_event_to_gdam_clicked),
			   NULL);
}

static void signal_name_hacks(const char* signal_name,
                              GtkObject*  object)
{
	if (strcmp(signal_name, "gdam_right_clicked") == 0
	 || strcmp(signal_name, "gdam-right-clicked") == 0
	 || strcmp(signal_name, "gdam_middle_clicked") == 0
	 || strcmp(signal_name, "gdam-middle-clicked") == 0
	 || strcmp(signal_name, "gdam_left_clicked") == 0
	 || strcmp(signal_name, "gdam-left-clicked") == 0
	 || strcmp(signal_name, "gdam_control_clicked") == 0
	 || strcmp(signal_name, "gdam-control-clicked") == 0) {
		if (left_clicked_id == 0) {
			left_clicked_id =
				gtk_object_class_user_signal_new
					(gtk_type_class(GTK_TYPE_WIDGET),
					 "gdam-left-clicked",
					 GTK_RUN_LAST,
					 gtk_marshal_NONE__NONE,
					 GTK_TYPE_NONE,
					 0);
			middle_clicked_id =
				gtk_object_class_user_signal_new
					(gtk_type_class(GTK_TYPE_WIDGET),
					 "gdam-middle-clicked",
					 GTK_RUN_LAST,
					 gtk_marshal_NONE__NONE,
					 GTK_TYPE_NONE,
					 0);
			right_clicked_id =
				gtk_object_class_user_signal_new
					(gtk_type_class(GTK_TYPE_WIDGET),
					 "gdam-right-clicked",
					 GTK_RUN_LAST,
					 gtk_marshal_NONE__NONE,
					 GTK_TYPE_NONE,
					 0);
			control_clicked_id =
				gtk_object_class_user_signal_new
					(gtk_type_class(GTK_TYPE_WIDGET),
					 "gdam-control-clicked",
					 GTK_RUN_LAST,
					 gtk_marshal_NONE__NONE,
					 GTK_TYPE_NONE,
					 0);
		}
		attach_simulated_clicks(GTK_WIDGET(object));
		return;
	}
}
