/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamskinutil.h"
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkrange.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkprogress.h>
#include <gdam/gdamdebug.h>
#include "gdamdnd.h"


GtkAdjustment* util_get_widget_adjustment(GtkWidget* widget)
{
	if (GTK_IS_SPIN_BUTTON(widget)) {
		return gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(widget));
	}
	if (GTK_IS_RANGE(widget)) {
		return gtk_range_get_adjustment(GTK_RANGE(widget));
	}
	if (GTK_IS_PROGRESS(widget)) {
		return GTK_PROGRESS(widget)->adjustment;
	}
	g_error("no adjustment in widget %s",
	        gtk_type_name(GTK_OBJECT(widget)->klass->type));
	return NULL;
}

/* TODO: actually do the caching. */
GdkColor* util_alloc_color_cached(int r, int g, int b)
{
	GdkColor* rv = g_new(GdkColor, 1);
	rv->red = r;
	rv->green = g;
	rv->blue = b;
	gdk_color_alloc(gdk_colormap_get_system(), rv);
	return rv;
}

typedef struct _DragHookInfo{
	GdamStringDragFunc	func;
	gpointer		user_data;
	char*			data_type;
	GtkTargetEntry*		entry;
} DragHookInfo;

typedef struct _DropHookInfo{
	GdamStringDropFunc	func;
	gpointer		user_data;
	char*			data_type;
	GtkTargetEntry*		entry;
} DropHookInfo;

static void dh_info_free(DropHookInfo* info)
{
	g_free(info->data_type);
	g_free(info->entry);
	g_free(info);
}

enum {
    TARGET_STRING
};

static void on_string_drag(
		GtkWidget 		*widget, 
		GdkDragContext 		*context, 
		GtkSelectionData 	*selection_data, 
		guint 			info, 
		guint32 		clk_time, 
		gpointer 		data) 
{
	DragHookInfo* dhinfo = (DragHookInfo*)data;	
	gdam_debug("on_string_drag... selection_data=%s, expected %s",
			gdk_atom_name(selection_data->target),
			dhinfo->data_type);
	if (selection_data->target == gdk_atom_intern(dhinfo->data_type, TRUE))
		(*dhinfo->func)(selection_data, dhinfo->user_data);
}

static void on_string_drop(
		GtkObject*        object,
		GdkDragContext*   context,
		gint              x,
		gint              y,
		GtkSelectionData* sdata, 
		guint             info,
		guint             clk_time,
		gpointer          data)
{
	DropHookInfo* dhinfo = (DropHookInfo*) data;
	if     (((dhinfo->data_type 
	     	&& sdata->type == gdk_atom_intern(dhinfo->data_type,    TRUE)))
				&& sdata->length >= 0 
				&& sdata->format == 8){
		    /* TODO: emit_full? */
		gdam_debug("valid type dropped");	
		dhinfo->func((char*)sdata->data, dhinfo->user_data);
	}
	else {
		gdam_debug("invalid type dropped");	
	}
    	gtk_drag_finish (context, TRUE, FALSE, clk_time);
}

void gdam_widget_string_source_set(GtkWidget* 		widget, 
				   GdamStringDragFunc 	func,
				   const char* 		o_data_type,
				   gpointer		user_data)
{
	DragHookInfo* dhinfo = g_new0(DragHookInfo, 1);
	GtkTargetEntry* entry = g_new(GtkTargetEntry, 1);
	char* data_type = g_strdup(o_data_type);
	entry[0].target = data_type;
	entry[0].flags = 0;
	entry[0].info = TARGET_STRING;
	dhinfo->func = func;
	dhinfo->user_data = user_data;
	dhinfo->data_type = data_type;
	dhinfo->entry = entry;
	gtk_drag_source_set(widget, 
			    gdam_drag_mask, 
			    entry,
			    1, 
			    GDK_ACTION_COPY);
	gtk_signal_connect_full(GTK_OBJECT(widget), 
			        "drag_data_get", 
			        on_string_drag, 
				NULL,
			        dhinfo,
			        (GDestroyNotify) dh_info_free,
			        FALSE,
			        FALSE);
}

void gdam_widget_string_dest_set  (GtkWidget* 		widget, 
				   GdamStringDropFunc 	func,
				   const char* 		o_data_type,
				   gpointer		user_data)
{
	DropHookInfo* dhinfo = g_new0(DropHookInfo, 1);
	GtkTargetEntry* entry = g_new(GtkTargetEntry, 1);
	char* data_type = g_strdup(o_data_type);
	entry[0].target = data_type;
	entry[0].flags = 0; 
	entry[0].info = TARGET_STRING;
	dhinfo->func = func;
	dhinfo->user_data = user_data;
	dhinfo->data_type = data_type;
	dhinfo->entry = entry;
	gtk_drag_dest_set(widget,
			  GTK_DEST_DEFAULT_ALL, 
			  entry,
			  1,
			  GDK_ACTION_COPY);
	gtk_signal_connect_full(GTK_OBJECT(widget),
			        "drag_data_received",
			        on_string_drop,
				NULL,
			        dhinfo,
				(GDestroyNotify) dh_info_free,
				FALSE,
				FALSE);
	return;
}

char** gdam_strdupv(char** input)
{
	int len = 0;
	char** at, **rv;
	at = input;
	while (*at) {
		at++;
		len++;
	}
	at = input;
	rv = g_new(char*, len + 1);
	len = 0;
	while (*at)
		rv[len++] = g_strdup(*at++);
	rv[len] = NULL;
	return rv;
}



void gdam_set_container(GtkContainer* new_container, GtkWidget* child)
{
	gtk_widget_ref(child);
	if (child->parent)
		gtk_container_remove(GTK_CONTAINER(child->parent), child);
	gtk_container_add(new_container, child);
	gtk_widget_unref(child);
}
	
static gboolean init_adj(gpointer data)
{
	GtkAdjustment* adj = GTK_ADJUSTMENT(data);
	gtk_adjustment_value_changed(adj);
	return FALSE;  /* don't be called again */
}

void gdam_delayed_adjustment_init	  (GtkAdjustment*	adj)
{
	g_idle_add(init_adj, adj);
}

static void add_to_kill_list(GtkWidget* widget, gpointer ptr)
{
	GSList** plist = ptr;
	*plist = g_slist_prepend(*plist, widget);
}

static void remove_from_container(gpointer widget, gpointer container)
{
	gtk_container_remove(GTK_CONTAINER(container), GTK_WIDGET(widget));
}

void gdam_gtk_container_remove_all(GtkContainer* container)
{
	GSList* list = NULL;
	gtk_container_foreach(container, add_to_kill_list, &list);
	list = g_slist_reverse(list);
	g_slist_foreach(list, remove_from_container, container);
	g_slist_free(list);
}
